#include "cl.hpp"

#include <iostream>
#include <fstream>
#include <streambuf>
#include <time.h>
#include <string.h>

using namespace cl;
using namespace std;

#define checkError(func) \
  if (errcode != CL_SUCCESS)\
  {\
    cout << "Error in " #func "\nError code = " << errcode << "\n";\
    exit(1);\
  }

#define checkErrorEx(command) \
  command; \
  checkError(command);

double getTime(Event & event)
{
  cl_ulong time_start, time_end;
  cl_int errcode;
  errcode = event.getProfilingInfo<cl_ulong>(CL_PROFILING_COMMAND_START, &time_start);
  errcode = event.getProfilingInfo<cl_ulong>(CL_PROFILING_COMMAND_END, &time_end);
  checkError(event.getEventProfilingInfo);
  return (double)(time_end - time_start)/1e9;
}

void check(float* host_result, float* host_result_check, int N)
{
  // check
  for (int i = 0; i < N; i++)
      if (abs(host_result[i] - host_result_check[i]) > 1e-6)
      {
          cout << "Error in element N " << i << ": result[i] = " << host_result[i] << " result_check[i] = " << host_result_check[i] << "\n";
          exit(1);
      }
}
  
int main()
{
  int device_index = 0;
  cl_int errcode;
  
  int N = 1*1000*1000, M = 128;
  float *host_x, *host_result, *host_result_check;
  host_x = new float[N];
  host_result = new float[N];
  host_result_check = new float[N];
  for (int i = 0; i < N; i++) host_x[i] = i;
  clock_t startCPU = clock();
//#pragma omp parallel for
  for (int i = 0; i<N; i++)
  {
    float loc_x = host_x[i];
    float sum=.0f;
    for (int j=0; j<M; j++)
      sum += 1.0f/(loc_x*loc_x+j);
    host_result_check[i] = sum;
  }
  double elapsedTimeCPU = (double)(clock()-startCPU)/CLOCKS_PER_SEC;
  
  //код kernel-функций
  ifstream file("kernel.c");
  if (!file) { cout << "Error: kernel.c not found\n"; exit(-1); }
  string sourceString((istreambuf_iterator<char>(file)), istreambuf_iterator<char>());
  
  //получаем список доступных OpenCL-платформ (драйверов OpenCL)
  vector<Platform> platform;//массив в который будут записываться идентификаторы платформ
  checkErrorEx( errcode = Platform::get(&platform) );
  cout << "OpenCL platforms found: " << platform.size() << "\n";
  cout << "Platform[0] is : " << platform[0].getInfo<CL_PLATFORM_VENDOR>() << " ver. " << platform[0].getInfo<CL_PLATFORM_VERSION>() << "\n";
  
  //в полученном списке платформ находим устройство GPU (видеокарту)
  vector<Device> devices;
  platform[0].getDevices(CL_DEVICE_TYPE_GPU, &devices);
  cout << "GPGPU devices found: " << devices.size() << "\n";
  if (devices.size() == 0)
  {
      cout << "Warning: YOU DON'T HAVE GPGPU. Then CPU will be used instead.\n";
      checkErrorEx( errcode = platform[0].getDevices(CL_DEVICE_TYPE_CPU, &devices) );
      cout << "CPU devices found: " << devices.size() << "\n";
      if (devices.size() == 0) {cout << "Error: CPU devices not found\n"; exit(-1);}
  }
  cout << "Use device N " << device_index << ": " << devices[device_index].getInfo<CL_DEVICE_NAME>() << "\n";
  
  //создаем контекст на видеокарте
  checkErrorEx( Context context(devices, NULL, NULL, NULL, &errcode) );
  
  //создаем очередь задач для контекста
  checkErrorEx( CommandQueue queue(context, devices[device_index], CL_QUEUE_PROFILING_ENABLE, &errcode) );// третий параметр - свойства
  
  //создаем обьект-программу с заданным текстом программы
  checkErrorEx( Program program = Program(context, sourceString, false/*build*/, &errcode) );  
  
  //компилируем и линкуем программу для видеокарты 
  errcode = program.build(devices, "-cl-fast-relaxed-math -cl-no-signed-zeros -cl-mad-enable");
  if (errcode != CL_SUCCESS)
  {
      cout << "There were error during build kernel code. Please, check program code. Errcode = " << errcode << "\n";
      cout << "BUILD LOG: " << program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(devices[device_index]) << "\n";
      return 1;
  }
  //создаем буфферы в видеопамяти
  checkErrorEx( Buffer dev_x = Buffer( context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, N*sizeof(float),  host_x, &errcode ) );
  checkErrorEx( Buffer dev_result = Buffer( context, CL_MEM_READ_WRITE, N*sizeof(float),  NULL, &errcode ) );
  
  //создаем объект - точку входа GPU-программы
  make_kernel< Buffer, Buffer, cl_uint, cl_uint> calc(program, "calc", &errcode);
  checkError(make_kernel);
  make_kernel< Buffer, Buffer, cl_uint, cl_uint> calc4(program, "calc4", &errcode);
  checkError(make_kernel);
  //создаем объект, соответствующий определенной конфигурации запуска kernel
  EnqueueArgs enqueueArgs(queue, NDRange(12*1024)/*globalSize*/, NullRange/*blockSize*/);
  
  //запускаем и ждем
  Event event = calc(enqueueArgs, dev_x, dev_result, M, N);
  checkErrorEx( errcode = event.wait() );
  double elapsedTimeGPU = getTime(event);
  checkErrorEx( errcode = queue.enqueueReadBuffer(dev_result, true, 0, N*sizeof(float), host_result, NULL, NULL) );
  check(host_result, host_result_check, N);
  
  event = calc4(enqueueArgs, dev_x, dev_result, M, N);
  checkErrorEx( errcode = event.wait() );
  double elapsedTimeGPU4 = getTime(event);
  checkErrorEx( errcode = queue.enqueueReadBuffer(dev_result, true, 0, N*sizeof(float), host_result, NULL, NULL) );
  check(host_result, host_result_check, N);
  
  cout << "CPU calc time = " << elapsedTimeCPU*1000 << " ms\n";
  cout << "GPU calc time = " << elapsedTimeGPU*1000 << " ms\n";
  cout << "GPU calc4 time = " << elapsedTimeGPU4*1000 << " ms\n";
  return 0;
}